@@ -1,17 +1,17 @@
 import json
 from pathlib import Path
-from schema import Schema, Use, Or, And, Optional
+from schema import Schema, Use
 from shutil import rmtree
 from typing import Union
 
+from SCAutolib import exceptions, schema_cas, schema_user
 from SCAutolib import (logger, run, LIB_DIR, LIB_BACKUP, LIB_DUMP,
                        LIB_DUMP_USERS, LIB_DUMP_CAS, LIB_DUMP_CARDS)
-from SCAutolib.exceptions import SCAutolibWrongConfig, SCAutolibException
 from SCAutolib.models import CA, file, user, card, authselect as auth
 from SCAutolib.models.file import File
 from SCAutolib.utils import (OSVersion, _check_selinux, _gen_private_key,
                              _get_os_version, _install_packages,
-                             _check_packages, dump_to_json)
+                             _check_packages, dump_to_json, local_ca_factory)
 
 
 class Controller:
@@ -54,9 +54,13 @@ class Controller:
         with self._lib_conf_path.open("r") as f:
             tmp_conf = json.load(f)
             if tmp_conf is None:
-                raise SCAutolibException("Data are not loaded correctly.")
+                raise exceptions.SCAutolibException(
+                    "Data are not loaded correctly.")
         self.lib_conf = self._validate_configuration(tmp_conf, params)
         self.users = []
+        for d in (LIB_DIR, LIB_BACKUP, LIB_DUMP, LIB_DUMP_USERS, LIB_DUMP_CAS,
+                  LIB_DUMP_CARDS):
+            d.mkdir(exist_ok=True)
 
     def prepare(self, force: bool, gdm: bool, install_missing: bool):
         """
@@ -68,11 +72,11 @@ class Controller:
         # In this method not having section for local CA and/or for IPA CA is OK
         try:
             self.setup_local_ca(force=force)
-        except SCAutolibWrongConfig as e:
+        except exceptions.SCAutolibWrongConfig as e:
             logger.info(e)
         try:
             self.setup_ipa_client(force=force)
-        except SCAutolibWrongConfig as e:
+        except exceptions.SCAutolibWrongConfig as e:
             logger.info(e)
         for usr in self.lib_conf["users"]:
             u = self.setup_user(usr, force=force)
@@ -117,7 +121,7 @@ class Controller:
             msg = "Can't continue. Some packages are missing: " \
                   f"{', '.join(missing)}"
             logger.critical(msg)
-            raise SCAutolibException(msg)
+            raise exceptions.SCAutolibException(msg)
 
         run(['dnf', 'groupinstall', "Smart Card Support", '-y'])
         logger.debug("Smart Card Support group in installed.")
@@ -125,8 +129,12 @@ class Controller:
         self.sssd_conf.create()
         self.sssd_conf.save()
         self._general_steps_for_virtual_sc()
-        # FIXME: think how to do this better
-        run(["useradd", "base-user"], return_code=[0, 9])
+
+        base_user = user.BaseUser("base-user", "redhat")
+        base_user.add_user()
+        dump_to_json(base_user)
+        dump_to_json(user.BaseUser(username="root",
+                                   password=self.lib_conf["root_passwd"]))
 
     def setup_local_ca(self, force: bool = False):
         """
@@ -142,24 +150,10 @@ class Controller:
 
         if "local_ca" not in self.lib_conf["ca"]:
             msg = "Section for local CA is not found in the configuration file"
-            raise SCAutolibWrongConfig(msg)
+            raise exceptions.SCAutolibWrongConfig(msg)
 
         ca_dir: Path = self.lib_conf["ca"]["local_ca"]["dir"]
-        cnf = file.OpensslCnf(ca_dir.joinpath("ca.cnf"), "CA", str(ca_dir))
-        self.local_ca = CA.LocalCA(root_dir=ca_dir, cnf=cnf)
-
-        if force:
-            logger.warning(f"Removing previous local CA in a directory "
-                           f"{ca_dir}")
-            self.local_ca.cleanup()
-        if not ca_dir.exists():
-            ca_dir.mkdir(exist_ok=True)
-
-            cnf.create()
-            cnf.save()
-            self.local_ca.setup()
-
-        run(["systemctl", "restart", "sssd"])
+        self.local_ca = local_ca_factory(ca_dir, force)
         logger.info(f"Local CA is configured in {ca_dir}")
 
         dump_to_json(self.local_ca)
@@ -177,7 +171,7 @@ class Controller:
         """
         if "ipa" not in self.lib_conf["ca"]:
             msg = "Section for IPA is not found in the configuration file"
-            raise SCAutolibWrongConfig(msg)
+            raise exceptions.SCAutolibWrongConfig(msg)
         self.ipa_ca = CA.IPAServerCA(**self.lib_conf["ca"]["ipa"])
 
         if self.ipa_ca.is_installed:
@@ -214,8 +208,6 @@ class Controller:
         if force and card_dir.exists():
             rmtree(card_dir)
 
-        user_dict["card_dir"].mkdir(exist_ok=True, parents=True)
-
         if user_dict["local"]:
             new_user = user.User(username=user_dict["name"],
                                  pin=user_dict["pin"],
@@ -224,13 +216,15 @@ class Controller:
                                  cert=user_dict["cert"], key=user_dict["key"],
                                  local=True)
             if force:
+                if new_user.cert and new_user.cert.exists():
+                    self.local_ca.revoke_cert(new_user.cert)
                 new_user.delete_user()
-                user_dict["card_dir"].mkdir(exist_ok=True)
+            user_dict["card_dir"].mkdir(exist_ok=True)
+            new_user.add_user()
 
             csr_path = new_user.card_dir.joinpath(f"csr-{new_user.username}.csr")
             cnf = file.OpensslCnf(filepath=csr_path, conf_type="user",
                                   replace=new_user.username)
-            new_user.add_user()
             cnf.create()
             cnf.save()
 
@@ -240,13 +234,14 @@ class Controller:
                 key="matchrule",
                 value=f"<SUBJECT>.*CN={new_user.username}.*")
             self.sssd_conf.save()
+            run(["systemctl", "restart", "sssd"])
             logger.debug(f"Match rule for user {new_user.username} is added "
                          f"to /etc/sssd/sssd.conf")
         else:
             if self.ipa_ca is None:
                 msg = "Can't proceed in configuration of IPA user because no " \
                       "IPA Client is configured"
-                raise SCAutolibException(msg)
+                raise exceptions.SCAutolibException(msg)
             new_user = user.IPAUser(ipa_server=self.ipa_ca,
                                     username=user_dict["name"],
                                     pin=user_dict["pin"],
@@ -256,13 +251,13 @@ class Controller:
                                     key=user_dict["key"],
                                     local=False)
             if force:
+                if new_user.cert and new_user.cert.exists():
+                    self.ipa_ca.revoke_cert(new_user.cert)
                 new_user.delete_user()
-                user_dict["card_dir"].mkdir(exist_ok=True)
+            user_dict["card_dir"].mkdir(exist_ok=True)
 
             new_user.add_user()
 
-        new_card = None
-
         if user_dict["card_type"] == "virtual":
             hsm_conf = file.SoftHSM2Conf(
                 new_user.card_dir.joinpath("sofhtsm2.conf"),
@@ -270,48 +265,46 @@ class Controller:
             hsm_conf.create()
             hsm_conf.save()
 
-            new_card = card.VirtualCard(new_user)
-            new_card.softhsm2_conf = hsm_conf.path
+            new_user.card = card.VirtualCard(new_user,
+                                             softhsm2_conf=hsm_conf.path)
         else:
-            raise NotImplementedError("Other card type than 'virtual' does not "
-                                      "supported yet")
-        new_card.create()
-        new_user.card = new_card
+            raise NotImplementedError("Other card type than 'virtual' does"
+                                      "not supported yet")
+
+        new_user.card.create()
         self.users.append(new_user)
 
         dump_to_json(new_user)
 
         return new_user
 
-    def enroll_card(self, user_: user.User):
+    def enroll_card(self, user_: user.User, force: bool = False):
         """
         Enroll the card of a given user with configured CA. If private key
         and/or the certificate do not exists, new one's would be requested
         from corresponding CA.
 
+        :param force:
         :param user_: User with a card to be enrolled.
         """
         logger.debug(f"Starting enrollment of the card for user "
                      f"{user_.username}")
         if not user_.card:
-            raise SCAutolibException(f"Card for the user {user_.username} is "
-                                     f"not initialized")
-        if user_.cert is None:
+            raise exceptions.SCAutolibException(
+                f"Card for the user {user_.username} is not initialized")
+
+        if not user_.key.exists():
+            _gen_private_key(user_.key)
+
+        if not user_.cert.exists():
             # Creating a new private key makes sense only if the certificate
             # doesn't exist yet
-            if user_.key is None:
-                user_.key = user_.card_dir.joinpath("private.key")
-                _gen_private_key(user_.key)
 
             csr = user_.gen_csr()
-            if user_.cert is None:
-                user_.cert = user_.card_dir.joinpath(
-                    f"cert-{user_.username}.pem")
 
-            if isinstance(user_, user.IPAUser):
-                self.ipa_ca.request_cert(csr, user_.username, user_.cert)
-            else:
-                self.local_ca.request_cert(csr, user_.username, user_.cert)
+            ca = self.ipa_ca \
+                if isinstance(user_, user.IPAUser) else self.local_ca
+            user_.cert = ca.request_cert(csr, user_.username, user_.cert)
 
         user_.card.enroll()
 
@@ -360,32 +353,6 @@ class Controller:
         #  Add loading of the values from params dict
         # IP regex
         # Specify validation schema for CAs
-        schema_cas = Schema(And(
-            Use(dict),
-            # Check that CA section contains at least one and maximum
-            # two entries
-            lambda l: 1 <= len(l.keys()) <= 2,
-            {Optional("local_ca"): {
-                Optional("dir", default=Path("/etc/SCAutolib/ca")): Use(Path)},
-                Optional("ipa"): {
-                    'admin_passwd': Use(str),
-                    'root_passwd': Use(str),
-                    Optional('ip_addr', default=None): Use(str),
-                    'server_hostname': Use(str),
-                    'client_hostname': Use(str),
-                    'domain': Use(str),
-                    'realm': Use(str.upper)}}),
-            ignore_extra_keys=True)
-
-        # Specify validation schema for all users
-        schema_user = Schema({'name': Use(str),
-                              'passwd': Use(str),
-                              'pin': Use(str),
-                              Optional('card_dir', default=None): Use(Path),
-                              'card_type': Or("virtual", "real", "removinator"),
-                              Optional('cert', default=None): Use(Path),
-                              Optional('key', default=None): Use(Path),
-                              'local': Use(bool)})
 
         # Specify general schema for whole config file
         schema = Schema({"root_passwd": Use(str),
@@ -418,7 +385,7 @@ class Controller:
         opensc_module.backup()
         try:
             opensc_module.get("disable-in", separator=":")
-        except SCAutolibException:
+        except exceptions.SCAutolibException:
             logger.warning("OpenSC module does not have option 'disable-in: "
                            "virt_cacard' set")
             opensc_module.set(key="disable-in", value="virt_cacard",
@@ -453,3 +420,36 @@ class Controller:
             return ["freeipa-client"]
         else:
             return ["ipa-client"]
+
+    def get_user_dict(self, name):
+        """
+        Get user dictionary from the config file.
+
+        :param name: name of the user
+        :type name: str
+        :return: user dictionary
+        :rtype: dict
+        """
+        for user_dict in self.lib_conf["users"]:
+            if user_dict["name"] == name:
+                return user_dict
+        raise exceptions.SCAutolibMissingUserConfig(name)
+
+    def init_ca(self, local: bool = False):
+        """
+        Initialize CA.
+
+        :param local: if True, local CA is initialized, otherwise IPA
+        :type local: bool
+        """
+        if local:
+            self.local_ca = CA.LocalCA(self.lib_conf["ca"]["local_ca"]["dir"])
+            if not self.local_ca.cert.exists():
+                raise exceptions.SCAutolibMissingCA(
+                    f"CA certificate not found in {str(self.local_ca.cert)}")
+        else:
+            self.ipa_ca = CA.IPAServerCA(self.lib_conf["ca"]["ipa"])
+            if not self.ipa_ca.is_installed:
+                raise exceptions.SCAutolibMissingCA(
+                    "IPA server CA is not installed")
+        logger.info("CA is initialized")
